from collections import defaultdict, OrderedDict

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models, IntegrityError
from django.db.models.query import QuerySet
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from etc.toolbox import get_model_class_from_string

from .settings import MODEL_FLAG
from .utils import get_flag_model

USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


@python_2_unicode_compatible
class FlagBase(models.Model):
    """Base class for flag models.
    Flags are marks on various site entities (model instances).

    Inherit from this model and override SITEFLAGS_FLAG_MODEL in settings.py
    to customize model fields and behaviour.

    """
    note = models.TextField(_('Note'), blank=True)
    status = models.IntegerField(_('Status'), null=True, blank=True, db_index=True)

    user = models.ForeignKey(
        USER_MODEL, related_name='%(class)s_users', verbose_name=_('User'), on_delete=models.CASCADE)
    time_created = models.DateTimeField(_('Date created'), auto_now_add=True)

    # Here follows a link to an object.
    object_id = models.PositiveIntegerField(verbose_name=_('Object ID'), db_index=True)
    content_type = models.ForeignKey(
        ContentType, verbose_name=_('Content type'), related_name='%(app_label)s_%(class)s_flags',
        on_delete=models.CASCADE)

    linked_object = GenericForeignKey()

    class Meta(object):
        abstract = True
        verbose_name = _('Flag')
        verbose_name_plural = _('Flags')
        unique_together = ('content_type', 'object_id', 'user', 'status')

    @classmethod
    def get_flags_for_types(cls, mdl_classes, user=None, status=None, allow_empty=True):
        """Returns a dictionary with flag objects associated with the given model classes (types).
        The dictionary is indexed by model classes.
        Each dict entry contains a list of associated flag objects.

        :param list mdl_classes:
        :param User user:
        :param int status:
        :param bool allow_empty: Flag. Include results for all given types, even those without associated flags.
        :return:
        """
        if not mdl_classes or (user and not user.id):
            return {}

        types_for_models = ContentType.objects.get_for_models(*mdl_classes, for_concrete_models=False)
        filter_kwargs = {'content_type__in': types_for_models.values()}
        update_filter_dict(filter_kwargs, user, status)

        flags = cls.objects.filter(**filter_kwargs).order_by('-time_created')
        flags_dict = defaultdict(list)
        for flag in flags:
            flags_dict[flag.content_type_id].append(flag)

        result = OrderedDict()  # Respect initial order.
        for mdl_cls in mdl_classes:
            content_type_id = types_for_models[mdl_cls].id
            if content_type_id in flags_dict:
                result[mdl_cls] = flags_dict[content_type_id]
            elif allow_empty:
                result[mdl_cls] = tuple()
        return result

    @classmethod
    def get_flags_for_objects(cls, objects_list, user=None, status=None):
        """Returns a dictionary with flag objects associated with the given model objects.
        The dictionary is indexed by objects IDs.
        Each dict entry contains a list of associated flag objects.

        :param list, QuerySet objects_list:
        :param User user:
        :param int status:
        :return:
        """
        if not objects_list or (user and not user.id):
            return {}

        objects_ids = objects_list
        if not isinstance(objects_list, QuerySet):
            objects_ids = [obj.pk for obj in objects_list]

        filter_kwargs = {
            'object_id__in': objects_ids,
            # Consider this list homogeneous.
            'content_type': ContentType.objects.get_for_model(objects_list[0], for_concrete_model=False)
        }
        update_filter_dict(filter_kwargs, user, status)
        flags = cls.objects.filter(**filter_kwargs)
        flags_dict = defaultdict(list)
        for flag in flags:
            flags_dict[flag.object_id].append(flag)

        result = {}
        for obj in objects_list:
            if obj.pk in flags_dict:
                result[obj.pk] = flags_dict[obj.pk]
            else:
                result[obj.pk] = tuple()
        return result

    def __str__(self):
        return '%s:%s status %s' % (self.content_type, self.object_id, self.status)


class Flag(FlagBase):
    """Built-in flag class. Default functionality."""


class ModelWithFlag(models.Model):
    """Helper base class for models with flags.

    Inherit from this model to be able to mark model instances.

    """

    flags = GenericRelation(MODEL_FLAG)

    class Meta(object):
        abstract = True

    @classmethod
    def get_flags_for_types(cls, mdl_classes, user=None, status=None, allow_empty=True):
        """Returns a dictionary with flag objects associated with the given model classes (types).
        The dictionary is indexed by model classes.
        Each dict entry contains a list of associated flag objects.

        :param list mdl_classes:
        :param User user:
        :param int status:
        :param bool allow_empty: Flag. Include results for all given types, even those without associated flags.
        :return:
        """
        return get_model_class_from_string(MODEL_FLAG).get_flags_for_types(
            mdl_classes, user=user, status=status, allow_empty=allow_empty)

    @classmethod
    def get_flags_for_objects(cls, objects_list, user=None, status=None):
        """Returns a dictionary with flag objects associated with the given model objects.
        The dictionary is indexed by objects IDs.
        Each dict entry contains a list of associated flag objects.

        :param list, QuerySet objects_list:
        :param User user:
        :param int status:
        :return:
        """
        return get_model_class_from_string(MODEL_FLAG).get_flags_for_objects(objects_list, user=user, status=status)

    def get_flags(self, user=None, status=None):
        """Returns flags for the object optionally filtered by status.

        :param User user: Optional user filter
        :param int status: Optional status filter
        :return:
        """
        filter_kwargs = {}
        update_filter_dict(filter_kwargs, user, status)
        return self.flags.filter(**filter_kwargs).all()

    def set_flag(self, user, note=None, status=None):
        """Flags the object.

        :param User user:
        :param str note: User-defined note for this flag.
        :param int status: Optional status integer (the meaning is defined by a developer).
        :return:
        """
        if not user.id:
            return None

        init_kwargs = {
            'user': user,
            'linked_object': self,
        }
        if note is not None:
            init_kwargs['note'] = note
        if status is not None:
            init_kwargs['status'] = status

        flag = get_flag_model()(**init_kwargs)
        try:
            flag.save()
        except IntegrityError:  # Record already exists.
            pass
        return flag

    def remove_flag(self, user=None, status=None):
        """Removes flag(s) from the object.

        :param User user: Optional user filter
        :param int status: Optional status filter
        :return:
        """
        filter_kwargs = {
            'content_type': ContentType.objects.get_for_model(self),
            'object_id': self.id
        }
        update_filter_dict(filter_kwargs, user, status)
        get_flag_model().objects.filter(**filter_kwargs).delete()

    def is_flagged(self, user=None, status=None):
        """Returns boolean whether the objects is flagged by a user.

        :param User user: Optional user filter
        :param int status: Optional status filter
        :return:
        """
        filter_kwargs = {
            'content_type': ContentType.objects.get_for_model(self),
            'object_id': self.id,
        }
        update_filter_dict(filter_kwargs, user, status)
        return self.flags.filter(**filter_kwargs).count()


def update_filter_dict(d, user, status):
    """Helper. Updates filter dict for a queryset.

    :param dict d:
    :param User|None user:
    :param int|None status:
    :return:
    """
    if user is not None:
        if not user.id:
            return None
        d['user'] = user
    if status is not None:
        d['status'] = status



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class SiteflagsConfig(AppConfig):
    """Siteflags configuration."""

    name = 'siteflags'
    verbose_name = _('Site Flags')



from uuid import uuid4

import pytest

from siteflags.models import ModelWithFlag


@pytest.fixture
def create_comment():
    from siteflags.tests.testapp.models import Comment

    def create_comment_():
        comment = Comment(title='comment%s' % uuid4().hex)
        comment.save()
        return comment

    return create_comment_


@pytest.fixture
def create_article():
    from siteflags.tests.testapp.models import Article

    def create_article_():
        article = Article(title='article%s' % uuid4().hex)
        article.save()
        return article

    return create_article_


class TestModelWithFlag(object):
    
    def test_get_flags_for_types(self, user, user_create, create_comment, create_article):

        from siteflags.tests.testapp.models import Comment, Article

        user2 = user_create()

        article_1 = create_article()
        article_2 = create_article()
        article_1.set_flag(user)
        article_1.set_flag(user2)
        article_2.set_flag(user2, status=44)

        flags = ModelWithFlag.get_flags_for_types([Article, Comment])
        assert len(flags) == 2
        assert len(flags[Article]) == 3

        comment_1 = create_comment()
        comment_2 = create_comment()
        comment_1.set_flag(user2)
        comment_1.set_flag(user)
        comment_2.set_flag(user, status=44)

        flags = ModelWithFlag.get_flags_for_types([Article, Comment])
        assert len(flags) == 2
        assert len(flags[Article]) == 3
        assert len(flags[Comment]) == 3

    def test_get_flags_for_objects(self, user, user_create, create_article):
        user2 = user_create()

        article_1 = create_article()
        article_2 = create_article()
        article_3 = create_article()
        articles_list = (article_1, article_2, article_3)

        article_1.set_flag(user)
        article_1.set_flag(user2)
        article_2.set_flag(user2, status=33)

        flags = ModelWithFlag.get_flags_for_objects(articles_list)
        assert len(flags) == len(articles_list)
        assert len(flags[article_1.pk]) == 2
        assert len(flags[article_2.pk]) == 1
        assert len(flags[article_3.pk]) == 0

        flags = ModelWithFlag.get_flags_for_objects(articles_list, user=user)
        assert len(flags) == len(articles_list)
        assert len(flags[article_1.pk]) == 1
        assert len(flags[article_2.pk]) == 0
        assert len(flags[article_3.pk]) == 0

        flags = ModelWithFlag.get_flags_for_objects(articles_list, status=33)
        assert len(flags) == len(articles_list)
        assert len(flags[article_1.pk]) == 0
        assert len(flags[article_2.pk]) == 1
        assert len(flags[article_3.pk]) == 0

    def test_set_flag(self, user, create_article):

        flag = create_article().set_flag(user, note='anote', status=10)

        assert flag.user == user
        assert flag.note == 'anote'
        assert flag.status == 10
        
    def test_get_flags(self, user, user_create, create_article):
        article = create_article()

        for idx in range(1, 5):
            article.set_flag(user, status=idx)

        user2 = user_create()
        article.set_flag(user2, status=2)

        flags = article.get_flags()
        assert len(flags) == 5

        flags = article.get_flags(status=2)
        assert len(flags) == 2

    def test_is_flagged(self, user, user_create, create_article):
        article = create_article()
        assert not article.is_flagged()

        article.set_flag(user, status=11)
        assert article.is_flagged()

        user2 = user_create()

        assert article.is_flagged(user)
        assert not article.is_flagged(user2)

        assert not article.is_flagged(user, status=12)
        assert article.is_flagged(user, status=11)

    def test_remove_flag(self, user, user_create, create_article):
        article = create_article()
        article.set_flag(user, status=11)
        article.set_flag(user, status=7)
        article.set_flag(user, status=13)
        user2 = user_create()
        article.set_flag(user2, status=11)
        article.set_flag(user2, status=13)
        user3 = user_create()
        article.set_flag(user3, status=11)

        flags = article.get_flags()
        assert len(flags) == 6

        article.remove_flag(user3)
        flags = article.get_flags()
        assert len(flags) == 5
        flags = article.get_flags(user3)
        assert len(flags) == 0

        article.remove_flag(user, status=13)
        flags = article.get_flags(user)
        assert len(flags) == 2

        article.remove_flag(status=11)
        flags = article.get_flags()
        assert len(flags) == 2

        article.remove_flag()
        flags = article.get_flags()
        assert len(flags) == 0



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Flag',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('note', models.TextField(verbose_name='Note', blank=True)),
                ('status', models.IntegerField(db_index=True, null=True, verbose_name='Status', blank=True)),
                ('time_created', models.DateTimeField(auto_now_add=True, verbose_name='Date created')),
                ('object_id', models.PositiveIntegerField(verbose_name='Object ID', db_index=True)),
                ('content_type', models.ForeignKey(related_name='siteflags_flag_flags', verbose_name='Content type', to='contenttypes.ContentType', on_delete=models.CASCADE)),
                ('user', models.ForeignKey(related_name='flag_users', verbose_name='User', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Flag',
                'verbose_name_plural': 'Flags',
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='flag',
            unique_together=set([('content_type', 'object_id', 'user', 'status')]),
        ),
    ]



from etc.toolbox import get_model_class_from_settings

from siteflags import settings


def get_flag_model():
    """Returns the Flag model, set for the project."""
    return get_model_class_from_settings(settings, 'MODEL_FLAG')



from django.contrib import admin

from .utils import get_flag_model

FLAG_MODEL = get_flag_model()


class FlagModelAdmin(admin.ModelAdmin):
    list_display = ('time_created', 'content_type', 'object_id', 'status')
    search_fields = ('object_id', 'content_type', 'user')
    list_filter = ('time_created', 'status', 'content_type')
    ordering = ('-time_created',)
    date_hierarchy = 'time_created'
admin.site.register(FLAG_MODEL, FlagModelAdmin)



from django.db import models

from siteflags.models import ModelWithFlag


class Comment(ModelWithFlag):

    title = models.CharField('title', max_length=255)


class Article(ModelWithFlag):

    title = models.CharField('title', max_length=255)
